@@ -1,6 +1,6 @@
 /*
  *   stunnel       TLS offloading and load-balancing proxy
- *   Copyright (C) 1998-2019 Michal Trojnara <Michal.Trojnara@stunnel.org>
+ *   Copyright (C) 1998-2025 Michal Trojnara <Michal.Trojnara@stunnel.org>
  *
  *   This program is free software; you can redistribute it and/or modify it
  *   under the terms of the GNU General Public License as published by the
@@ -35,7 +35,6 @@
  *   forward this exception.
  */
 
-#include "common.h"
 #include "prototypes.h"
 
 #ifndef SHUT_RD
@@ -56,9 +55,11 @@ NOEXPORT void local_start(CLI *);
 NOEXPORT void remote_start(CLI *);
 NOEXPORT void ssl_start(CLI *);
 NOEXPORT void session_cache_retrieve(CLI *);
+#if OPENSSL_VERSION_NUMBER >= 0x10101000L
+NOEXPORT void print_tmp_key(SSL *s);
+#endif
 NOEXPORT void print_cipher(CLI *);
 NOEXPORT void transfer(CLI *);
-NOEXPORT int parse_socket_error(CLI *, const char *);
 
 NOEXPORT void auth_user(CLI *);
 NOEXPORT SOCKET connect_local(CLI *);
@@ -73,7 +74,7 @@ NOEXPORT void connect_setup(CLI *);
 NOEXPORT int connect_init(CLI *, int);
 NOEXPORT int redirect(CLI *);
 NOEXPORT void print_bound_address(CLI *);
-NOEXPORT void reset(SOCKET, char *);
+NOEXPORT void reset(SOCKET, const char *);
 
 /* allocate local data structure for the new thread */
 CLI *alloc_client_session(SERVICE_OPTIONS *opt, SOCKET rfd, SOCKET wfd) {
@@ -174,7 +175,7 @@ void ignore_value(void *ptr) {
 void client_main(CLI *c) {
     s_log(LOG_DEBUG, "Service [%s] started", c->opt->servname);
     if(c->opt->exec_name && c->opt->connect_addr.names) {
-        if(c->opt->option.retry)
+        if(c->opt->retry >= 0)
             exec_connect_loop(c);
         else
             exec_connect_once(c);
@@ -199,8 +200,8 @@ void client_free(CLI *c) {
 #endif /* __GNUC__ */
 NOEXPORT void exec_connect_loop(CLI *c) {
     unsigned long long seq=0;
-    char *fresh_id=c->tls->id;
-    unsigned retry;
+    const char *fresh_id=c->tls->id;
+    long retry;
 
     do {
         /* make sure c->tls->id is valid in str_printf() */
@@ -211,19 +212,20 @@ NOEXPORT void exec_connect_loop(CLI *c) {
         exec_connect_once(c);
         /* retry is asynchronously changed in the main thread,
          * so we make sure to use the same value for both checks */
-        retry=c->opt->option.retry;
-        if(retry) {
+        retry=c->opt->retry;
+        if(retry >= 0) {
             s_log(LOG_INFO, "Retrying an exec+connect section");
             /* c and id are detached, so it is safe to call str_stats() */
             str_stats(); /* client thread allocation tracking */
-            s_poll_sleep(1, 0);
+            if(retry)
+                s_poll_sleep((int)(retry/1000), (int)(retry%1000));
             c->rr++;
         }
 
         /* make sure c->tls->id is valid in str_free() */
         c->tls->id=fresh_id;
         str_free(id);
-    } while(retry); /* retry is disabled on config reload */
+    } while(retry >= 0); /* retry is disabled on config reload */
 }
 #ifdef __GNUC__
 #if __GNUC__ > 4 || (__GNUC__ == 4 && __GNUC_MINOR__ >= 6)
@@ -259,7 +261,8 @@ NOEXPORT void exec_connect_once(CLI *fresh_c) {
 #endif /* __GNUC__ */
 NOEXPORT void client_run(CLI *c) {
     jmp_buf exception_buffer, *exception_backup;
-    int err, rst;
+    int err, rst_remote, rst_local;
+    char *close_status;
 #ifndef USE_FORK
     int num;
 #endif
@@ -299,11 +302,29 @@ NOEXPORT void client_run(CLI *c) {
     }
     c->exception_pointer=exception_backup;
 
-    rst=err==1 && c->opt->option.reset;
+    /* plain socket is remote in the server mode */
+    rst_remote=err==1 && c->opt->option.reset &&
+        (!c->fatal_alert || !c->opt->option.client);
+    /* plain socket is local in the client mode */
+    rst_local=err==1 && c->opt->option.reset &&
+        (!c->fatal_alert || c->opt->option.client);
+    if(rst_remote && rst_local)
+        close_status="reset";
+    else if(rst_remote)
+        close_status="reset/closed";
+    else if(rst_local)
+        close_status="closed/reset";
+    else
+        close_status="closed";
     s_log(LOG_NOTICE,
         "Connection %s: %llu byte(s) sent to TLS, %llu byte(s) sent to socket",
-        rst ? "reset" : "closed",
+        close_status,
         (unsigned long long)c->ssl_bytes, (unsigned long long)c->sock_bytes);
+#ifdef USE_WIN32
+    if(capwin_hwnd && (rst_remote || rst_local)
+            && InterlockedExchange(&capwin_connectivity, 0))
+        PostMessage(capwin_hwnd, WM_CAPWIN_NET_DOWN, 0, 0);
+#endif
 
         /* cleanup temporary (e.g. IDENT) socket */
     if(c->fd!=INVALID_SOCKET)
@@ -332,8 +353,11 @@ NOEXPORT void client_run(CLI *c) {
 
         /* cleanup the remote socket */
     if(c->remote_fd.fd!=INVALID_SOCKET) { /* remote socket initialized */
-        if(rst && c->remote_fd.is_socket) /* reset */
-            reset(c->remote_fd.fd, "linger (remote)");
+        if(rst_remote && c->remote_fd.is_socket) /* reset */
+            reset(c->remote_fd.fd, "remote_fd");
+#ifndef USE_UCONTEXT
+        set_nonblock(c->remote_fd.fd, 0);
+#endif
         closesocket(c->remote_fd.fd);
         s_log(LOG_DEBUG, "Remote descriptor (FD=%ld) closed",
             (long)c->remote_fd.fd);
@@ -343,16 +367,19 @@ NOEXPORT void client_run(CLI *c) {
         /* cleanup the local socket */
     if(c->local_rfd.fd!=INVALID_SOCKET) { /* local socket initialized */
         if(c->local_rfd.fd==c->local_wfd.fd) {
-            if(rst && c->local_rfd.is_socket)
-                reset(c->local_rfd.fd, "linger (local)");
+            if(rst_local && c->local_rfd.is_socket)
+                reset(c->local_rfd.fd, "local_rfd/local_wfd");
+#ifndef USE_UCONTEXT
+            set_nonblock(c->local_rfd.fd, 0);
+#endif
             closesocket(c->local_rfd.fd);
             s_log(LOG_DEBUG, "Local descriptor (FD=%ld) closed",
                 (long)c->local_rfd.fd);
         } else { /* stdin/stdout */
-            if(rst && c->local_rfd.is_socket)
-                reset(c->local_rfd.fd, "linger (local_rfd)");
-            if(rst && c->local_wfd.is_socket)
-                reset(c->local_wfd.fd, "linger (local_wfd)");
+            if(rst_local && c->local_rfd.is_socket)
+                reset(c->local_rfd.fd, "local_rfd");
+            if(rst_local && c->local_wfd.is_socket)
+                reset(c->local_wfd.fd, "local_wfd");
         }
         c->local_rfd.fd=c->local_wfd.fd=INVALID_SOCKET;
     }
@@ -387,18 +414,23 @@ NOEXPORT void client_run(CLI *c) {
 #endif /* __GNUC__ */
 
 NOEXPORT void client_try(CLI *c) {
+    c->flag.redirect=0;
     local_start(c);
-    protocol(c, c->opt, PROTOCOL_EARLY);
+    if(c->opt->protocol_early)
+        c->opt->protocol_early(c);
     if(c->opt->option.connect_before_ssl) {
         remote_start(c);
-        protocol(c, c->opt, PROTOCOL_MIDDLE);
+        if(c->opt->protocol_middle)
+            c->opt->protocol_middle(c);
         ssl_start(c);
     } else {
         ssl_start(c);
-        protocol(c, c->opt, PROTOCOL_MIDDLE);
+        if(c->opt->protocol_middle && !c->flag.redirect)
+            c->opt->protocol_middle(c);
         remote_start(c);
     }
-    protocol(c, c->opt, PROTOCOL_LATE);
+    if(c->opt->protocol_late && !c->flag.redirect)
+        c->opt->protocol_late(c);
     transfer(c);
 }
 
@@ -410,15 +442,19 @@ NOEXPORT void local_start(CLI *c) {
     addr_len=sizeof(SOCKADDR_UNION);
     c->local_rfd.is_socket=!getpeername(c->local_rfd.fd, &addr.sa, &addr_len);
     if(c->local_rfd.is_socket) {
+        /* store the retrieved peer address */
         memcpy(&c->peer_addr.sa, &addr.sa, (size_t)addr_len);
         c->peer_addr_len=addr_len;
-        if(socket_options_set(c->opt, c->local_rfd.fd, 1))
+        if(
+#ifdef HAVE_STRUCT_SOCKADDR_UN
+                addr.sa.sa_family!=AF_UNIX &&
+#endif
+                socket_options_set(c->opt, c->local_rfd.fd, 1)) {
             s_log(LOG_WARNING, "Failed to set local socket options");
-    } else {
-        if(get_last_socket_error()!=S_ENOTSOCK) {
-            sockerror("getpeerbyname (local_rfd)");
-            throw_exception(c, 1);
         }
+    } else if(get_last_socket_error()!=S_ENOTSOCK) {
+        sockerror("getpeerbyname (local_rfd)");
+        throw_exception(c, 1);
     }
 
     /* check if local_wfd is a socket and get peer address */
@@ -428,17 +464,21 @@ NOEXPORT void local_start(CLI *c) {
         addr_len=sizeof(SOCKADDR_UNION);
         c->local_wfd.is_socket=!getpeername(c->local_wfd.fd, &addr.sa, &addr_len);
         if(c->local_wfd.is_socket) {
-            if(!c->local_rfd.is_socket) { /* already retrieved */
+            if(!c->local_rfd.is_socket) { /* not retrieved from rfd? */
+                /* store the retrieved peer address */
                 memcpy(&c->peer_addr.sa, &addr.sa, (size_t)addr_len);
                 c->peer_addr_len=addr_len;
             }
-            if(socket_options_set(c->opt, c->local_wfd.fd, 1))
+            if(
+#ifdef HAVE_STRUCT_SOCKADDR_UN
+                    addr.sa.sa_family!=AF_UNIX &&
+#endif
+                    socket_options_set(c->opt, c->local_wfd.fd, 1)) {
                 s_log(LOG_WARNING, "Failed to set local socket options");
-        } else {
-            if(get_last_socket_error()!=S_ENOTSOCK) {
-                sockerror("getpeerbyname (local_wfd)");
-                throw_exception(c, 1);
             }
+        } else if(get_last_socket_error()!=S_ENOTSOCK) {
+            sockerror("getpeerbyname (local_wfd)");
+            throw_exception(c, 1);
         }
     }
 
@@ -476,7 +516,7 @@ NOEXPORT void remote_start(CLI *c) {
         c->bind_addr=NULL; /* don't bind */
 
     /* setup c->remote_fd, now */
-    if(c->opt->exec_name && !c->opt->connect_addr.names && !redirect(c))
+    if(c->opt->exec_name && !c->opt->connect_addr.names && !c->flag.redirect)
         c->remote_fd.fd=connect_local(c); /* not for exec+connect targets */
     else
         c->remote_fd.fd=connect_remote(c);
@@ -504,23 +544,29 @@ NOEXPORT void ssl_start(CLI *c) {
 
     c->ssl=SSL_new(c->opt->ctx);
     if(!c->ssl) {
-        sslerror("SSL_new");
+        ssl_error(c, "SSL_new");
         throw_exception(c, 1);
     }
     /* for callbacks */
     if(!SSL_set_ex_data(c->ssl, index_ssl_cli, c)) {
-        sslerror("SSL_set_ex_data");
+        ssl_error(c, "SSL_set_ex_data");
         throw_exception(c, 1);
     }
     if(c->opt->option.client) {
 #ifndef OPENSSL_NO_TLSEXT
+#ifndef OPENSSL_NO_OCSP
+        if(!SSL_set_tlsext_status_type(c->ssl, TLSEXT_STATUSTYPE_ocsp)) {
+            ssl_error(c, "OCSP: SSL_set_tlsext_status_type");
+            throw_exception(c, 1);
+        }
+#endif /* !defined(OPENSSL_NO_OCSP) */
         /* c->opt->sni should always be initialized at this point,
          * either explicitly with "sni"
          * or implicitly with "protocolHost" or "connect" */
         if(c->opt->sni && *c->opt->sni) {
             s_log(LOG_INFO, "SNI: sending servername: %s", c->opt->sni);
             if(!SSL_set_tlsext_host_name(c->ssl, c->opt->sni)) {
-                sslerror("SSL_set_tlsext_host_name");
+                ssl_error(c, "SSL_set_tlsext_host_name");
                 throw_exception(c, 1);
             }
         } else { /* c->opt->sni was set to an empty value */
@@ -562,10 +608,7 @@ NOEXPORT void ssl_start(CLI *c) {
             CRYPTO_THREAD_write_lock(stunnel_locks[LOCK_SSL]);
 #endif /* OpenSSL version < 1.1.0 */
 
-        if(c->opt->option.client)
-            i=SSL_connect(c->ssl);
-        else
-            i=SSL_accept(c->ssl);
+        i=c->opt->option.client ? SSL_connect(c->ssl) : SSL_accept(c->ssl);
 
 #if OPENSSL_VERSION_NUMBER < 0x10100000L
         if(unsafe_openssl)
@@ -606,48 +649,136 @@ NOEXPORT void ssl_start(CLI *c) {
 #endif
                 continue;
             }
+            sockerror(c->opt->option.client ? "SSL_connect" : "SSL_accept");
+            throw_exception(c, 1);
         }
-        if(c->opt->option.client)
-            sslerror("SSL_connect");
-        else
-            sslerror("SSL_accept");
+        ssl_error(c, c->opt->option.client ? "SSL_connect" : "SSL_accept");
         throw_exception(c, 1);
     }
+    ERR_clear_error(); /* silence any cached errors */
     print_cipher(c);
     sess=SSL_get1_session(c->ssl);
     if(sess) {
+        X509 *peer_cert=SSL_get_peer_certificate(c->ssl);
+        if(peer_cert) {
+            X509_free(peer_cert);
+        } else { /* no authentication was performed */
+            if(!SSL_SESSION_set_ex_data(sess,
+                    index_session_authenticated, NULL)) {
+                ssl_error(c, "SSL_SESSION_set_ex_data");
+                SSL_SESSION_free(sess);
+                throw_exception(c, 1);
+            }
+        }
         if(SSL_session_reused(c->ssl)) {
-            print_session_id(sess);
+            /* otherwise printed from sess_new_cb() */
+            print_session_id("Session id", sess);
         } else { /* a new session was negotiated */
             /* SSL_SESS_CACHE_NO_INTERNAL_STORE prevented automatic caching */
             if(!c->opt->option.client)
                 SSL_CTX_add_session(c->opt->ctx, sess);
         }
         SSL_SESSION_free(sess);
+    } else if(c->opt->redirect_addr.names) {
+        s_log(LOG_ERR, "No session available for redirection");
+        throw_exception(c, 1);
     }
+    c->flag.redirect=(unsigned)redirect(c)&1;
 }
 
 NOEXPORT void session_cache_retrieve(CLI *c) {
-    SSL_SESSION *sess;
+    SSL_SESSION *sess=NULL;
 
     CRYPTO_THREAD_read_lock(stunnel_locks[LOCK_SESSION]);
-    if(c->opt->option.delayed_lookup) {
-        sess=c->opt->session;
-    } else { /* per-destination client cache */
-        if(c->opt->connect_session) {
-            sess=c->opt->connect_session[c->idx];
-        } else {
-            s_log(LOG_ERR, "INTERNAL ERROR: Uninitialized client session cache");
-            sess=NULL;
-        }
-    }
+    if(c->opt->connect_session) /* per-destination client session */
+        sess=c->opt->connect_session[c->idx];
+    if(!sess) /* either no per-destination cache or no session cached */
+        sess=c->opt->session; /* fallback session */
     if(sess)
         SSL_set_session(c->ssl, sess);
     CRYPTO_THREAD_unlock(stunnel_locks[LOCK_SESSION]);
+
+    if(sess)
+        print_session_id("Attempting to resume", sess);
+    else
+        s_log(LOG_DEBUG, "No previous session to resume");
 }
 
+#if OPENSSL_VERSION_NUMBER >= 0x10101000L
+NOEXPORT void print_tmp_key(SSL *s) {
+    EVP_PKEY *key;
+    long tmp_key_found;
+
+#ifdef SSL_CTRL_GET_PEER_TMP_KEY
+    tmp_key_found=SSL_get_peer_tmp_key(s, &key);
+#else
+    tmp_key_found=SSL_get_server_tmp_key(s, &key);
+#endif
+    if(!tmp_key_found) {
+        s_log(LOG_INFO, "No peer temporary key received");
+#if OPENSSL_VERSION_NUMBER>=0x30000000L
+        if(SSL_version(s) == TLS1_3_VERSION) {
+            s_log(LOG_INFO, "Negotiated TLSv1.3 group: %s",
+                SSL_group_to_name(s, (int)SSL_get_negotiated_group(s)));
+        }
+#endif /* OPENSSL_VERSION_NUMBER>=0x30000000L */
+        return;
+    }
+    switch(EVP_PKEY_id(key)) {
+#if OPENSSL_VERSION_NUMBER>=0x30000000L
+    case EVP_PKEY_KEYMGMT:
+        {
+            const char *keyname=EVP_PKEY_get0_type_name(key);
+
+            if(keyname)
+                s_log(LOG_INFO, "Peer temporary key: %s, %d bits",
+                    keyname, EVP_PKEY_bits(key));
+            else
+                s_log(LOG_INFO, "Unable to determine temporary key type");
+        }
+        break;
+#endif /* OPENSSL_VERSION_NUMBER>=0x30000000L */
+    case EVP_PKEY_RSA:
+        s_log(LOG_INFO, "Peer temporary key: RSA, %d bits", EVP_PKEY_bits(key));
+        break;
+    case EVP_PKEY_DH:
+        s_log(LOG_INFO, "Peer temporary key: DH, %d bits", EVP_PKEY_bits(key));
+        break;
+#ifndef OPENSSL_NO_EC
+    case EVP_PKEY_EC:
+        {
+            EC_KEY *ec=EVP_PKEY_get1_EC_KEY(key);
+            int nid=EC_GROUP_get_curve_name(EC_KEY_get0_group(ec));
+            const char *cname=EC_curve_nid2nist(nid);
+
+            EC_KEY_free(ec);
+            if (cname == NULL)
+                cname=OBJ_nid2sn(nid);
+            s_log(LOG_INFO, "Peer temporary key: ECDH, %s, %d bits", cname, EVP_PKEY_bits(key));
+        }
+        break;
+#endif
+    default:
+        {
+            int keyid=EVP_PKEY_id(key);
+
+            if(keyid != NID_undef)
+                s_log(LOG_INFO, "Peer temporary key: %s, %d bits",
+                    OBJ_nid2sn(keyid), EVP_PKEY_bits(key));
+            else
+                s_log(LOG_INFO, "Unable to determine temporary key type");
+        }
+    }
+    EVP_PKEY_free(key);
+}
+#endif /* OpenSSL 1.1.1 or later */
+
 NOEXPORT void print_cipher(CLI *c) { /* print negotiated cipher */
+#if OPENSSL_VERSION_NUMBER >= 0x10101000L
+    const SSL_CIPHER *cipher;
+#else
     SSL_CIPHER *cipher;
+#endif /* OpenSSL 1.1.1 or later */
 #ifndef OPENSSL_NO_COMP
     const COMP_METHOD *compression, *expansion;
 #endif
@@ -655,15 +786,27 @@ NOEXPORT void print_cipher(CLI *c) { /* print negotiated cipher */
     if(c->opt->log_level<LOG_INFO) /* performance optimization */
         return;
 
-    s_log(LOG_INFO, "TLS %s: %s",
-        c->opt->option.client ? "connected" : "accepted",
-        SSL_session_reused(c->ssl) && !c->flag.psk ?
-            "previous session reused" : "new session negotiated");
+#ifndef OPENSSL_NO_PSK
+    if(c->flag.psk_found) {
+        if(c->opt->option.client) {
+            s_log(LOG_ERR, "INTERNAL ERROR: PSK found on a client");
+        } else {
+            s_log(LOG_INFO, "TLS accepted: PSK");
+        }
+    } else
+#endif /* !defined(OPENSSL_NO_PSK) */
+        s_log(LOG_INFO, "TLS %s: %s",
+            c->opt->option.client ? "connected" : "accepted",
+            SSL_session_reused(c->ssl) ?
+                "previous session reused" : "new session negotiated");
 
     cipher=(SSL_CIPHER *)SSL_get_current_cipher(c->ssl);
     s_log(LOG_INFO, "%s ciphersuite: %s (%d-bit encryption)",
         SSL_get_version(c->ssl), SSL_CIPHER_get_name(cipher),
         SSL_CIPHER_get_bits(cipher, NULL));
+#if OPENSSL_VERSION_NUMBER >= 0x10101000L
+    print_tmp_key(c->ssl);
+#endif
 
 #ifndef OPENSSL_NO_COMP
     compression=SSL_get_current_compression(c->ssl);
@@ -683,7 +826,6 @@ NOEXPORT void transfer(CLI *c) {
     int has_pending=0, prev_has_pending;
 #endif
     int watchdog=0; /* a counter to detect an infinite loop */
-    ssize_t num;
     int err;
     /* logical channels (not file descriptors!) open for read or write */
     int sock_open_rd=1, sock_open_wr=1;
@@ -710,7 +852,7 @@ NOEXPORT void transfer(CLI *c) {
 
         /****************************** setup c->fds structure */
         s_poll_init(c->fds, 0); /* initialize the structure */
-        /* for plain socket open data strem = open file descriptor */
+        /* for plain socket open data stream = open file descriptor */
         /* make sure to add each open socket to receive exceptions! */
         if(sock_open_rd) /* only poll if the read file descriptor is open */
             s_poll_add(c->fds, c->sock_rfd->fd, c->sock_ptr<BUFFSIZE, 0);
@@ -815,17 +957,21 @@ NOEXPORT void transfer(CLI *c) {
 
         /****************************** hangups without read or write */
         if(!(sock_can_rd || sock_can_wr || ssl_can_rd || ssl_can_wr)) {
-            if(s_poll_hup(c->fds, c->sock_rfd->fd) ||
-                    s_poll_hup(c->fds, c->sock_wfd->fd)) {
+            if(s_poll_hup(c->fds, c->sock_wfd->fd)) {
                 if(c->ssl_ptr) {
                     s_log(LOG_ERR,
-                        "Socket closed (HUP) with %ld unsent byte(s)",
+                        "Write socket closed (HUP) with %ld unsent byte(s)",
                         (long)c->ssl_ptr);
                     throw_exception(c, 1); /* reset the sockets */
                 }
-                s_log(LOG_INFO, "Socket closed (HUP)");
-                sock_open_rd=sock_open_wr=0;
-            } else if(s_poll_hup(c->fds, c->ssl_rfd->fd) ||
+                s_log(LOG_INFO, "Write socket closed (HUP)");
+                sock_open_wr=0;
+            }
+            if(s_poll_hup(c->fds, c->sock_rfd->fd)) {
+                s_log(LOG_INFO, "Read socket closed (HUP)");
+                sock_open_rd=0;
+            }
+            if(s_poll_hup(c->fds, c->ssl_rfd->fd) ||
                     s_poll_hup(c->fds, c->ssl_wfd->fd)) {
                 if(c->sock_ptr) {
                     s_log(LOG_ERR,
@@ -846,9 +992,9 @@ NOEXPORT void transfer(CLI *c) {
 
         /****************************** send TLS close_notify alert */
         if(shutdown_wants_read || shutdown_wants_write) {
-            num=SSL_shutdown(c->ssl); /* send close_notify alert */
+            int num=SSL_shutdown(c->ssl); /* send close_notify alert */
             if(num<0) /* -1 - not completed */
-                err=SSL_get_error(c->ssl, (int)num);
+                err=SSL_get_error(c->ssl, num);
             else /* 0 or 1 - success */
                 err=SSL_ERROR_NONE;
             switch(err) {
@@ -856,16 +1002,6 @@ NOEXPORT void transfer(CLI *c) {
                 s_log(LOG_INFO, "SSL_shutdown successfully sent close_notify alert");
                 shutdown_wants_read=shutdown_wants_write=0;
                 break;
-            case SSL_ERROR_SYSCALL: /* socket error */
-                if(parse_socket_error(c, "SSL_shutdown"))
-                    break; /* a non-critical error: retry */
-                SSL_set_shutdown(c->ssl, SSL_SENT_SHUTDOWN|SSL_RECEIVED_SHUTDOWN);
-                shutdown_wants_read=shutdown_wants_write=0;
-                break;
-            case SSL_ERROR_ZERO_RETURN: /* connection closed */
-                SSL_set_shutdown(c->ssl, SSL_SENT_SHUTDOWN|SSL_RECEIVED_SHUTDOWN);
-                shutdown_wants_read=shutdown_wants_write=0;
-                break;
             case SSL_ERROR_WANT_WRITE:
                 s_log(LOG_DEBUG, "SSL_shutdown returned WANT_WRITE: retrying");
                 shutdown_wants_read=0;
@@ -877,8 +1013,18 @@ NOEXPORT void transfer(CLI *c) {
                 shutdown_wants_write=0;
                 break;
             case SSL_ERROR_SSL: /* TLS error */
-                sslerror("SSL_shutdown");
+                ssl_error(c, "SSL_shutdown");
                 throw_exception(c, 1);
+            case SSL_ERROR_ZERO_RETURN: /* received a close_notify alert */
+                SSL_set_shutdown(c->ssl, SSL_SENT_SHUTDOWN|SSL_RECEIVED_SHUTDOWN);
+                shutdown_wants_read=shutdown_wants_write=0;
+                break;
+            case SSL_ERROR_SYSCALL: /* socket error */
+                if(socket_needs_retry(c, "transfer: SSL_shutdown"))
+                    break; /* a non-critical error: retry */
+                SSL_set_shutdown(c->ssl, SSL_SENT_SHUTDOWN|SSL_RECEIVED_SHUTDOWN);
+                shutdown_wants_read=shutdown_wants_write=0;
+                break;
             default:
                 s_log(LOG_ERR, "SSL_shutdown/SSL_get_error returned %d", err);
                 throw_exception(c, 1);
@@ -887,10 +1033,10 @@ NOEXPORT void transfer(CLI *c) {
 
         /****************************** write to socket */
         if(sock_open_wr && sock_can_wr) {
-            num=writesocket(c->sock_wfd->fd, c->ssl_buff, c->ssl_ptr);
+            ssize_t num=writesocket(c->sock_wfd->fd, c->ssl_buff, c->ssl_ptr);
             switch(num) {
             case -1: /* error */
-                if(parse_socket_error(c, "writesocket"))
+                if(socket_needs_retry(c, "transfer: writesocket"))
                     break; /* a non-critical error: retry */
                 sock_open_rd=sock_open_wr=0;
                 break;
@@ -908,11 +1054,11 @@ NOEXPORT void transfer(CLI *c) {
 
         /****************************** read from socket */
         if(sock_open_rd && sock_can_rd) {
-            num=readsocket(c->sock_rfd->fd,
+            ssize_t num=readsocket(c->sock_rfd->fd,
                 c->sock_buff+c->sock_ptr, BUFFSIZE-c->sock_ptr);
             switch(num) {
             case -1:
-                if(parse_socket_error(c, "readsocket"))
+                if(socket_needs_retry(c, "transfer: readsocket"))
                     break; /* a non-critical error: retry */
                 sock_open_rd=sock_open_wr=0;
                 break;
@@ -936,21 +1082,21 @@ NOEXPORT void transfer(CLI *c) {
         /****************************** write to TLS */
         if((write_wants_read && ssl_can_rd) ||
                 (write_wants_write && ssl_can_wr)) {
+            int num=SSL_write(c->ssl, c->sock_buff, (int)(c->sock_ptr));
             write_wants_read=0;
             write_wants_write=0;
-            num=SSL_write(c->ssl, c->sock_buff, (int)(c->sock_ptr));
-            switch(err=SSL_get_error(c->ssl, (int)num)) {
+            switch(err=SSL_get_error(c->ssl, num)) {
             case SSL_ERROR_NONE:
                 if(num==0) { /* nothing was written: ignore */
                     s_log(LOG_DEBUG, "SSL_write returned 0");
-                    break; /* do not reset the watchdog */
+                } else {
+                    memmove(c->sock_buff, c->sock_buff+num,
+                        c->sock_ptr-(size_t)num);
+                    c->sock_ptr-=(size_t)num;
+                    memset(c->sock_buff+c->sock_ptr, 0, (size_t)num); /* PPD */
+                    c->ssl_bytes+=(size_t)num;
+                    watchdog=0; /* data transferred -> reset the watchdog */
                 }
-                memmove(c->sock_buff, c->sock_buff+num,
-                    c->sock_ptr-(size_t)num);
-                c->sock_ptr-=(size_t)num;
-                memset(c->sock_buff+c->sock_ptr, 0, (size_t)num); /* paranoia */
-                c->ssl_bytes+=(size_t)num;
-                watchdog=0; /* reset the watchdog */
                 break;
             case SSL_ERROR_WANT_WRITE: /* buffered data? */
                 s_log(LOG_DEBUG, "SSL_write returned WANT_WRITE: retrying");
@@ -964,8 +1110,13 @@ NOEXPORT void transfer(CLI *c) {
                 s_log(LOG_DEBUG,
                     "SSL_write returned WANT_X509_LOOKUP: retrying");
                 break;
+            case SSL_ERROR_SSL:
+                ssl_error(c, "SSL_write");
+                throw_exception(c, 1);
+            case SSL_ERROR_ZERO_RETURN: /* a buffered close_notify alert */
+                /* fall through */
             case SSL_ERROR_SYSCALL: /* socket error */
-                if(num && parse_socket_error(c, "SSL_write"))
+                if(socket_needs_retry(c, "transfer: SSL_write") && num)
                     break; /* a non-critical error: retry */
                 /* EOF -> buggy (e.g. Microsoft) peer:
                  * TLS socket closed without close_notify alert */
@@ -978,14 +1129,6 @@ NOEXPORT void transfer(CLI *c) {
                 s_log(LOG_INFO, "TLS socket closed (SSL_write)");
                 SSL_set_shutdown(c->ssl, SSL_SENT_SHUTDOWN|SSL_RECEIVED_SHUTDOWN);
                 break;
-            case SSL_ERROR_ZERO_RETURN: /* close_notify alert received */
-                s_log(LOG_INFO, "TLS closed (SSL_write)");
-                if(SSL_version(c->ssl)==SSL2_VERSION)
-                    SSL_set_shutdown(c->ssl, SSL_SENT_SHUTDOWN|SSL_RECEIVED_SHUTDOWN);
-                break;
-            case SSL_ERROR_SSL:
-                sslerror("SSL_write");
-                throw_exception(c, 1);
             default:
                 s_log(LOG_ERR, "SSL_write/SSL_get_error returned %d", err);
                 throw_exception(c, 1);
@@ -997,17 +1140,17 @@ NOEXPORT void transfer(CLI *c) {
                 /* it may be possible to read some pending data after
                  * writesocket() above made some room in c->ssl_buff */
                 (read_wants_write && ssl_can_wr)) {
+            int num=SSL_read(c->ssl, c->ssl_buff+c->ssl_ptr, (int)(BUFFSIZE-c->ssl_ptr));
             read_wants_read=0;
             read_wants_write=0;
-            num=SSL_read(c->ssl, c->ssl_buff+c->ssl_ptr, (int)(BUFFSIZE-c->ssl_ptr));
-            switch(err=SSL_get_error(c->ssl, (int)num)) {
+            switch(err=SSL_get_error(c->ssl, num)) {
             case SSL_ERROR_NONE:
                 if(num==0) { /* nothing was read: ignore */
                     s_log(LOG_DEBUG, "SSL_read returned 0");
-                    break; /* do not reset the watchdog */
+                } else {
+                    c->ssl_ptr+=(size_t)num;
+                    watchdog=0; /* data transferred -> reset the watchdog */
                 }
-                c->ssl_ptr+=(size_t)num;
-                watchdog=0; /* reset the watchdog */
                 break;
             case SSL_ERROR_WANT_WRITE:
                 s_log(LOG_DEBUG, "SSL_read returned WANT_WRITE: retrying");
@@ -1023,8 +1166,35 @@ NOEXPORT void transfer(CLI *c) {
                 s_log(LOG_DEBUG,
                     "SSL_read returned WANT_X509_LOOKUP: retrying");
                 break;
+            case SSL_ERROR_SSL:
+#ifdef SSL_R_UNEXPECTED_EOF_WHILE_READING
+                /* OpenSSL 3.0 changed the method of reporting socket EOF */
+                if(ERR_GET_REASON(ERR_peek_error())==
+                        SSL_R_UNEXPECTED_EOF_WHILE_READING) {
+                    /* EOF -> buggy (e.g. Microsoft) peer:
+                    * TLS socket closed without close_notify alert */
+                    if(c->sock_ptr || write_wants_write) {
+                        s_log(LOG_ERR,
+                            "TLS socket closed (SSL_read) with %ld unsent byte(s)",
+                            (long)c->sock_ptr);
+                        throw_exception(c, 1); /* reset the sockets */
+                    }
+                    s_log(LOG_INFO, "TLS socket closed (SSL_read)");
+                    SSL_set_shutdown(c->ssl,
+                        SSL_SENT_SHUTDOWN|SSL_RECEIVED_SHUTDOWN);
+                    break;
+                }
+#endif /* SSL_R_UNEXPECTED_EOF_WHILE_READING */
+                ssl_error(c, "SSL_read");
+                throw_exception(c, 1);
+            case SSL_ERROR_ZERO_RETURN: /* received a close_notify alert */
+                s_log(LOG_INFO, "TLS closed (SSL_read)");
+                if(SSL_version(c->ssl)==SSL2_VERSION)
+                    SSL_set_shutdown(c->ssl,
+                        SSL_SENT_SHUTDOWN|SSL_RECEIVED_SHUTDOWN);
+                break;
             case SSL_ERROR_SYSCALL:
-                if(num && parse_socket_error(c, "SSL_read"))
+                if(socket_needs_retry(c, "transfer: SSL_read") && num)
                     break; /* a non-critical error: retry */
                 /* EOF -> buggy (e.g. Microsoft) peer:
                  * TLS socket closed without close_notify alert */
@@ -1038,15 +1208,6 @@ NOEXPORT void transfer(CLI *c) {
                 SSL_set_shutdown(c->ssl,
                     SSL_SENT_SHUTDOWN|SSL_RECEIVED_SHUTDOWN);
                 break;
-            case SSL_ERROR_ZERO_RETURN: /* close_notify alert received */
-                s_log(LOG_INFO, "TLS closed (SSL_read)");
-                if(SSL_version(c->ssl)==SSL2_VERSION)
-                    SSL_set_shutdown(c->ssl,
-                        SSL_SENT_SHUTDOWN|SSL_RECEIVED_SHUTDOWN);
-                break;
-            case SSL_ERROR_SSL:
-                sslerror("SSL_read");
-                throw_exception(c, 1);
             default:
                 s_log(LOG_ERR, "SSL_read/SSL_get_error returned %d", err);
                 throw_exception(c, 1);
@@ -1165,61 +1326,6 @@ NOEXPORT void transfer(CLI *c) {
         shutdown_wants_read || shutdown_wants_write);
 }
 
-#ifdef __GNUC__
-#if __GNUC__ > 4 || (__GNUC__ == 4 && __GNUC_MINOR__ >= 6)
-#pragma GCC diagnostic push
-#endif /* __GNUC__>=4.6 */
-#if __GNUC__ >= 7
-#pragma GCC diagnostic ignored "-Wimplicit-fallthrough"
-#endif /* __GNUC__>=7 */
-#endif /* __GNUC__ */
-
-    /* returns 0 on close and 1 on non-critical errors */
-NOEXPORT int parse_socket_error(CLI *c, const char *text) {
-    switch(get_last_socket_error()) {
-        /* http://tangentsoft.net/wskfaq/articles/bsd-compatibility.html */
-    case 0: /* close on read, or close on write on WIN32 */
-#ifndef USE_WIN32
-    case EPIPE: /* close on write on Unix */
-#endif
-    case S_ECONNABORTED:
-        s_log(LOG_INFO, "%s: Socket is closed", text);
-        return 0;
-    case S_EINTR:
-        s_log(LOG_DEBUG, "%s: Interrupted by a signal: retrying", text);
-        return 1;
-    case S_EWOULDBLOCK:
-        s_log(LOG_NOTICE, "%s: Would block: retrying", text);
-        s_poll_sleep(1, 0); /* Microsoft bug KB177346 */
-        return 1;
-#if S_EAGAIN!=S_EWOULDBLOCK
-    case S_EAGAIN:
-        s_log(LOG_DEBUG,
-            "%s: Temporary lack of resources: retrying", text);
-        return 1;
-#endif
-#ifdef USE_WIN32
-    case S_ECONNRESET:
-        /* dying "exec" processes on Win32 cause reset instead of close */
-        if(c->opt->exec_name) {
-            s_log(LOG_INFO, "%s: Socket is closed (exec)", text);
-            return 0;
-        }
-        /* fall through */
-#endif
-    default:
-        sockerror(text);
-        throw_exception(c, 1);
-        return -1; /* some C compilers require a return value */
-    }
-}
-
-#ifdef __GNUC__
-#if __GNUC__ > 4 || (__GNUC__ == 4 && __GNUC_MINOR__ >= 6)
-#pragma GCC diagnostic pop
-#endif /* __GNUC__>=4.6 */
-#endif /* __GNUC__ */
-
 NOEXPORT void auth_user(CLI *c) {
 #ifndef _WIN32_WCE
     struct servent *s_ent;    /* structure for getservbyname */
@@ -1251,7 +1357,7 @@ NOEXPORT void auth_user(CLI *c) {
         s_log(LOG_WARNING, "Unknown service 'auth': using default 113");
         ident.in.sin_port=htons(113);
     }
-    if(s_connect(c, &ident, addr_len(&ident)))
+    if(s_connect(c, &ident, addr_len(&ident), c->opt->timeout_connect))
         throw_exception(c, 1);
     s_log(LOG_DEBUG, "IDENT server connected");
     remote_port=ntohs(c->peer_addr.in.sin_port);
@@ -1290,7 +1396,7 @@ NOEXPORT void auth_user(CLI *c) {
     while(*user==' ') /* skip leading spaces */
         ++user;
     if(strcmp(user, c->opt->username)) {
-        s_log(LOG_WARNING, "Connection from %s REFUSED by IDENT (user \"%s\")",
+        s_log(LOG_ERR, "Connection from %s REFUSED by IDENT (user \"%s\")",
             c->accepted_address, user);
         str_free(line);
         throw_exception(c, 1);
@@ -1338,8 +1444,6 @@ NOEXPORT SOCKET connect_local(CLI *c) { /* spawn local process */
 
 #else /* standard Unix version */
 
-extern char **environ;
-
 NOEXPORT SOCKET connect_local(CLI *c) { /* spawn local process */
     int fd[2], pid;
     char **env;
@@ -1401,6 +1505,7 @@ NOEXPORT SOCKET connect_local(CLI *c) { /* spawn local process */
 }
 
 char **env_alloc(CLI *c) {
+    extern char **environ;
     char **env=NULL, **p;
     unsigned n=0; /* (n+2) keeps the list NULL-terminated */
     char *name, host[40], port[6];
@@ -1422,7 +1527,8 @@ char **env_alloc(CLI *c) {
             env[n++]=str_dup("LD_PRELOAD_32=" LIBDIR "/libstunnel.so");
             env=str_realloc(env, (n+2)*sizeof(char *));
             env[n++]=str_dup("LD_PRELOAD_64=" LIBDIR "/" MACH64 "/libstunnel.so");
-#elif __osf /* for Tru64 _RLD_LIST is used instead */
+#elif defined(__osf) || defined(__osf__)
+            /* for Tru64 _RLD_LIST is used instead */
             env=str_realloc(env, (n+2)*sizeof(char *));
             env[n++]=str_dup("_RLD_LIST=" LIBDIR "/libstunnel.so:DEFAULT");
 #else
@@ -1487,7 +1593,8 @@ NOEXPORT SOCKET connect_remote(CLI *c) {
         c->idx=(idx_start+idx_try)%c->connect_addr.num;
         if(!connect_init(c, c->connect_addr.addr[c->idx].sa.sa_family) &&
                 !s_connect(c, &c->connect_addr.addr[c->idx],
-                    addr_len(&c->connect_addr.addr[c->idx]))) {
+                    addr_len(&c->connect_addr.addr[c->idx]),
+                    c->opt->timeout_connect)) {
             if(c->ssl) {
                 SSL_SESSION *sess=SSL_get1_session(c->ssl);
                 if(sess) {
@@ -1532,7 +1639,7 @@ NOEXPORT void idx_cache_save(SSL_SESSION *sess, SOCKADDR_UNION *cur_addr) {
     if(ok) {
         str_free(old_addr); /* NULL pointers are ignored */
     } else { /* failed to store new_addr -> remove it */
-        sslerror("SSL_SESSION_set_ex_data");
+        ssl_error(NULL, "SSL_SESSION_set_ex_data");
         str_free(new_addr); /* NULL pointers are ignored */
     }
 }
@@ -1586,7 +1693,7 @@ NOEXPORT unsigned idx_cache_retrieve(CLI *c) {
 }
 
 NOEXPORT void connect_setup(CLI *c) {
-    if(redirect(c)) { /* process "redirect" first */
+    if(c->flag.redirect) { /* process "redirect" first */
         s_log(LOG_NOTICE, "Redirecting connection");
         /* c->connect_addr.addr may be allocated in protocol negotiations */
         str_free(c->connect_addr.addr);
@@ -1617,28 +1724,16 @@ NOEXPORT int connect_init(CLI *c, int domain) {
 
     if(c->bind_addr) {
         /* setup bind_addr based on c->bind_addr */
-#if 0
-        /* IPv6 addresses converted from IPv4 cause timeouts */
-#ifdef USE_IPv6
-        if(c->bind_addr->sa.sa_family==AF_INET && domain==AF_INET6) {
-            /* convert the binding address from IPv4 to IPv6 */
-            memset(&bind_addr, 0, sizeof bind_addr);
-            bind_addr.in6.sin6_family=AF_INET6;
-            bind_addr.in6.sin6_port=c->bind_addr->in.sin_port;
-            /* address format example: ::ffff:192.0.2.128 */
-            memset((char *)&bind_addr.in6.sin6_addr+10, 0xff, 2);
-            memcpy((char *)&bind_addr.in6.sin6_addr+12,
-                &c->bind_addr->in.sin_addr, 4);
-        } else /* just make a local copy */
-#endif
-#endif
-            memcpy(&bind_addr, c->bind_addr, (size_t)addr_len(c->bind_addr));
+        memcpy(&bind_addr, c->bind_addr, (size_t)addr_len(c->bind_addr));
         /* perform the initial sanity checks before creating a socket */
         if(bind_addr.sa.sa_family!=domain) {
             s_log(LOG_DEBUG, "Cannot assign an AF=%d address an AF=%d socket",
                 bind_addr.sa.sa_family, domain);
             return 1; /* failure */
         }
+    } else {
+        /* only needed to avoid a warning in MSVC */
+        memset(&bind_addr, 0, sizeof bind_addr);
     }
 
     /* create a new socket */
@@ -1745,13 +1840,18 @@ NOEXPORT void print_bound_address(CLI *c) {
     str_free(txt);
 }
 
-NOEXPORT void reset(SOCKET fd, char *txt) { /* set lingering on a socket */
+NOEXPORT void reset(SOCKET fd, const char *txt) { /* set lingering on a socket */
     struct linger l;
 
+    s_log(LOG_DEBUG, "%s reset (FD=%ld)", txt, (long)fd);
     l.l_onoff=1;
     l.l_linger=0;
-    if(setsockopt(fd, SOL_SOCKET, SO_LINGER, (void *)&l, sizeof l))
-        log_error(LOG_DEBUG, get_last_socket_error(), txt);
+    if(setsockopt(fd, SOL_SOCKET, SO_LINGER, (void *)&l, sizeof l)) {
+        int err=get_last_socket_error();
+        char *message=str_printf("setsockopt(SO_LINGER) on %s", txt);
+        log_error(LOG_INFO, err, message);
+        str_free(message);
+    }
 }
 
 void throw_exception(CLI *c, int v) {
